import { getChangedRanges, findChildrenInRange, createStyleTag, findChildren } from '@tiptap/core';
import CodeBlock from '@tiptap/extension-code-block';
import { Plugin, PluginKey } from '@tiptap/pm/state';
import { Decoration, DecorationSet } from '@tiptap/pm/view';
import { createHighlighterCoreSync, createJavaScriptRegexEngine } from 'shiki/core';
import css from 'shiki/langs/css.mjs';
import html from 'shiki/langs/html.mjs';
import javascript from 'shiki/langs/javascript.mjs';
import json from 'shiki/langs/json.mjs';
import markdown from 'shiki/langs/markdown.mjs';
import python from 'shiki/langs/python.mjs';
import typescript from 'shiki/langs/typescript.mjs';
import vue from 'shiki/langs/vue.mjs';
import vitesseDark from 'shiki/themes/vitesse-dark.mjs';
import vitesseLight from 'shiki/themes/vitesse-light.mjs';

const style = `
.tiptap .shiki {
  border-radius: 4px;
  padding: 8px;
}
.tiptap .shiki:focus-visible {
  outline: 2px solid transparent;
  outline-offset: 2px;
}

html.dark .tiptap .shiki,
html.dark .tiptap .shiki span {
  color: var(--shiki-dark) !important;
  background-color: var(--shiki-dark-bg) !important;
  font-style: var(--shiki-dark-font-style) !important;
  font-weight: var(--shiki-dark-font-weight) !important;
  text-decoration: var(--shiki-dark-text-decoration) !important;
}
`;

const codeBlock = CodeBlock.extend({
  addOptions() {
    return {
      ...this.parent?.(),
      languages: [css, html, javascript, json, markdown, python, typescript, vue],
      themes: {
        light: vitesseLight,
        dark: vitesseDark
      }
    };
  },
  addStorage() {
    const highlighter = createHighlighterCoreSync({
      langs: this.options.languages,
      themes: Object.values(this.options.themes),
      engine: createJavaScriptRegexEngine()
    });
    const languages = Array.from(new Set(this.options.languages.map((language) => language.map((item) => ({
      name: item.name,
      displayName: item.displayName
    })).filter((item) => !!item.displayName)).flat(2))).reduce((languages2, item) => {
      if (languages2.find((language) => language.name === item.name)) {
        return languages2;
      }
      languages2.push(item);
      return languages2;
    }, []);
    return {
      highlighter,
      languages
    };
  },
  addProseMirrorPlugins() {
    const getDecorations = (doc) => {
      const decorations = [];
      findChildren(doc, (node) => node.type.name === this.name).forEach(({ node, pos }) => {
        const language = node.attrs.language ?? "text";
        const hast = this.storage.highlighter.codeToHast(node.textContent, {
          lang: language,
          themes: this.options.themes
        });
        const preNode = hast.children[0];
        decorations.push(Decoration.node(pos, pos + node.nodeSize, preNode.properties));
        let from = pos + 1;
        const lines = preNode.children[0].children;
        for (const line of lines) {
          if (line.children?.length) {
            let lineFrom = from;
            line.children?.forEach((node2) => {
              const nodeLen = node2.children[0].value.length;
              decorations.push(Decoration.inline(lineFrom, lineFrom + nodeLen, node2.properties));
              lineFrom += nodeLen;
            });
            from = lineFrom;
          } else if (line.type === "text") {
            from += line.value.length;
          }
        }
      });
      return DecorationSet.create(doc, decorations);
    };
    return [
      new Plugin({
        key: new PluginKey(this.name),
        state: {
          init(_, { doc }) {
            return getDecorations(doc);
          },
          apply: (tr, value, _, { doc: newDoc }) => {
            if (!tr.docChanged) {
              return value.map(tr.mapping, tr.doc);
            }
            const changedRanges = getChangedRanges(tr);
            let hasChangedCodeBlock = false;
            for (const range of changedRanges) {
              hasChangedCodeBlock = !!findChildrenInRange(newDoc, range.newRange, (node) => node.type.name === this.name).length;
              if (hasChangedCodeBlock) {
                return getDecorations(newDoc);
              }
            }
            return value.map(tr.mapping, tr.doc);
          }
        },
        props: {
          decorations(state) {
            return this.getState(state);
          }
        }
      })
    ];
  },
  onBeforeCreate() {
    createStyleTag(style, void 0, this.name);
  }
});

export { codeBlock };
